/*

    Emphasized wave merger source code

    This code won't work for files larger than 2gb, as standard (non 64-bit)
    seek&tell functions are used

    george@oss3d.com

    experimental - trying to make wavemerge endian independent

*/


#ifndef CORE_WAVEREAD16_H_INCLUDED
#define CORE_WAVEREAD16_H_INCLUDED

#include <qstring.h>
#include <math.h>
#include "numtypes.h"

#define ftell64 ftell

#define PAD_END        256
#define PAD_FACTOR     ((CLUSTERSIZE/4)*2)


#pragma pack(push,1)

class WaveRead16
{
public:

    int data_offset;

    int rate,nch,nbits,nbytes,sz,loop_start,loop_sz;

    char *ibuf;
    int ibuf_sz;

    QFile f;

    struct head
    {
        int32_LE sz;
        int16_LE fmt;
        int16_LE nch;
        int32_LE sr;
        char reserved[6];
        int16_LE nbits;
    };

    struct SampleLoop 
    {
        uint32_LE  dwIdentifier;  /* link to 'cue' or 'adtl' entry */
        uint32_LE  dwType;        /* 0=fwd 1=alt 2=rev 3=... */
        uint32_LE  dwStart;       /* counting from 0 */
        uint32_LE  dwEnd;
        uint32_LE  dwFraction;    /* 0xFFFFFFFF = 1 sample point */
        uint32_LE  dwPlayCount;   /* 0=infinite */
    };

    struct SMPL
    {
        int32_LE  dwManufacturer;        /* assigned by MMA */
        int32_LE  dwProduct;             /* assigned by manufacturer */
        int32_LE  dwSamplePeriod;        /* nanoseconds  */
        int32_LE  dwMIDIUnityNote;
        int32_LE  dwMIDIPitchFraction;   /* 0xFFFFFFFF = +100 cents  */
        int32_LE  dwSMPTEFormat;         /* fps (or 29 for 30-drop) */
        int32_LE  dwSMPTEOffset;         /* 0xhhmmssff */
        int32_LE  cSampleLoops;          /* number of loops      */
        int32_LE  cbSamplerData;         /* bytes manufacturer specific data */
        struct SampleLoop Loops[1];
    };


    WaveRead16() {ibuf=0;ibuf_sz=262144;}
    ~WaveRead16() {if (f.isOpen()) f.close();if (ibuf) delete ibuf;}

    int open(const QString &fname)
    {
        if (f.isOpen()) f.close();

        f.setFileName(fname);
        if (!f.open(QIODevice::ReadOnly)) return -1;

        char tagbuf[9];
        int32_LE fsz;
        memset(tagbuf,0,sizeof(tagbuf));
        f.read(tagbuf,4);
        if (strcmp(tagbuf,"RIFF")!=0) {f.close();return -2;}
        f.read((char *)&fsz,4);
        fsz=fsz+4;
        f.read(tagbuf,8);
        if (strcmp(tagbuf,"WAVEfmt ")!=0) {f.close();return -2;}

        head hd;
        f.read((char*)&hd,sizeof(hd));
        f.seek(f.pos()+hd.sz-sizeof(hd)+4);

        nch   = hd.nch;
        rate  = hd.sr;
        nbits = hd.nbits;
        nbytes = nbits/8;

        loop_sz=-1;

        for(;;)
        {
            int32_LE size;
            int cseek;
            tagbuf[4]=0;
            if (f.read((char *)tagbuf,4)!=4) break;
            if (f.read((char *)&size,4)!=4) break;

            cseek=f.pos();

            if (cseek>=fsz) break;

            if (strcmp(tagbuf,"data")==0)
            {
                data_offset=cseek;
                if (nbits==16) sz=size/(nch*2);else
                if (nbits==24) sz=size/(nch*3);else
                {f.close();return -3;}

                if (loop_sz==-1)
                {
                    loop_start=sz;
                    loop_sz=0;
                }
            }else
            if (strcmp(tagbuf,"smpl")==0)
            {
                SMPL smpl;
                memset(&smpl,0,sizeof(smpl));
                f.read((char *)&smpl,size);         
                if (smpl.cSampleLoops>0)
                {
                    int q=smpl.Loops[0].dwEnd-smpl.Loops[0].dwStart+1;
                    sz=smpl.Loops[0].dwEnd+1;
                    loop_sz=q;
                    loop_start=sz-loop_sz;
                }
            }
            if (!f.seek(cseek+size)) break;            
            if (f.atEnd()) break;
        }
        return 1;
    }

    int shortread(int offs,int size,short *dest)
    {
        if (!ibuf) ibuf=new char[ibuf_sz];

        if ((size+offs)>=sz) size=sz-offs;

        int bytesize=nbytes*size*nch;

        if (bytesize>ibuf_sz)
        {
            size = ibuf_sz/(nch*nbytes);
            bytesize = nbytes*size*nch;    
        }

        if ((size<=0)||(bytesize<=0)) return size;

        f.seek(data_offset + offs*nbytes*nch);

        f.read((char *)ibuf,bytesize);

        if (nbits==16)
        {
            signed short *smp = (signed short *)ibuf;
            int t,tl=size*nch;
            for(t=0;t<tl;t++)
            {
                *(dest++) = smp[0];
                smp++;
            }
        }else
        if (nbits==24)
        {
            unsigned char *smp = (unsigned char *)ibuf;
            int t,tl=size*nch;
            for(t=0;t<tl;t++)
            {
                ((char *)dest)[0] = smp[1];
                ((char *)dest)[1] = smp[2];

                dest++;
                smp++;smp++;smp++;
            }
        }

        return size;
    }

    int read(int offs,int size,short *dest)
    {
        int rdsize,realsz=0;
        while(size)
        {
            rdsize = shortread(offs,size,dest);
            dest += nch*rdsize;
            size-=rdsize;
            offs+=rdsize;
            realsz+=rdsize;
            if (!rdsize) break;
        }
        return realsz;
    }

    int read2(int offs,int size,short *dest)
    {
        int rdsize,realsz=0;
        while(size)
        {
            rdsize = read(offs,size,dest);
            dest += nch*rdsize;
            size-=rdsize;
            realsz+=rdsize;
            if (!rdsize) break;
        }
        return realsz;
    }

};

#pragma pack(pop)

#endif // CORE_WAVEMERGE_H_INCLUDED
